/*********************************************************************/
/* wiczenie 9 - Obsuga interfejsu 1-wire - odczyt pastylki DS1990A */
/*               Obsuga wywietlacza LCD 2x16                       */
/* J.D. '2003                                                        */
/*********************************************************************/

#include <io2313.h>
#include <progmem.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

//  Ponisze definicje su do realizacji wygodnego dostpu bitowego
typedef struct _bit_struct
{
 unsigned char bit0: 1;
 unsigned char bit1: 1;
 unsigned char bit2: 1;
 unsigned char bit3: 1;
 unsigned char bit4: 1;
 unsigned char bit5: 1;
 unsigned char bit6: 1;
 unsigned char bit7: 1;
}pole_bitowe;

#define DAJ_BIT(adr) (*((volatile pole_bitowe*) (adr)))
#define _PORTB 0x38
#define _PINB 0x36
#define _PORTD 0x32
#define _DDRD 0x31
#define _PIND 0x30
#define pastylka_we DAJ_BIT(_PIND).bit3
#define pastylka_wy DAJ_BIT(_DDRD).bit3
#define lcd_rs DAJ_BIT(_PORTB).bit2
#define lcd_e DAJ_BIT(_PORTB).bit3
#define led0 DAJ_BIT(_PORTB).bit0

#define stan_0 1   //definicja stanu niskiego na linii1-wire
                   //"1" oznacza przeczenie portu w tryb wyjciowy
                   //port jest wczeniej wysterowany w stan niski
#define stan_1 0   //definicja stanu wysokiego na linii1-wire
                   //"0" oznacza przeczenie portu w tryb wejciowy
                   //stan wysoki jest wymuszany przez zewntrzny
                   //rezystor podcigajcy
#define CR 0x0a    //definicja znaku CR (przecie do nowej linii)

char buflcd[4];    //roboczy bufor wywietlacza LCD
char *pbuflcd;     //wskanik na bufor wywietlacza
unsigned char bufor1w[8];    //bufor interfejsu 1-wire (dane z pastylki)
unsigned char *pbufor1w;     //wskanik na bufor danych z pastylki
unsigned char wiersz=0;      //pozycja umieszczenia znaku na LCD
unsigned char kolumna=0;     //pozycja umieszczenia znaku na LCD

void czekaj(unsigned long zt) //procedura wytracania czasu
{
 #define tau 10.38
 unsigned char zt1;
 for(;zt>0;zt--)
 {
  for(zt1=255;zt1!=0;zt1--);
 }
}

//>>>>>>>>>>>>>    Procedury obsugi wywietlacza LCD   <<<<<<<<<<<<<<<<<<
void piszilcd(unsigned char instr)     //zapisz instrukcj sterujc do LCD
{
 lcd_rs=0;         
 lcd_e=1;
 PORTB=(PORTB&0x0f)|(instr&0xf0);  //przygotuj starszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD ok. 100us
 lcd_e=1;
 PORTB=(PORTB&0x0f)|((instr&0x0f)<<4); //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD ok. 100us
}

void piszdlcd(char dana)     //zapisz dan do LCD
{
 lcd_rs=1;
 lcd_e=1;
 PORTB=(PORTB&0x0f)|(dana&0xf0);  //przygotuj starszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD
 lcd_e=1;
 PORTB=(PORTB&0x0f)|((dana&0x0f)<<4);  //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD
}

void czysclcd(void)          //czy ekran
{
 piszilcd(0x01);   //polecenie czyszczenia ekranu dla kontrolera LCD
 czekaj(1.64*tau); //rozkaz 0x01 wykonuje si 1.64ms
 wiersz=0;
 kolumna=0;
}

void lcdxy(unsigned char w, unsigned char k)     //ustaw wsprzdne kursora
{
 piszilcd((w*0x40+k)|0x80);  //standardowy rozkaz sterownika LCD
}                            //ustawiajcy kursor w okrelonych wsprzdnych

void piszznak(char znak)     //procedura umieszcza znak na wywietlaczu
{
 piszdlcd(znak);             //wywietl znak na LCD
}

void pisztekst(char *tekst)  //pisz tekst na LCD wskazywany pointerem *tekst
{
 char zn;
 char nr=0;

 while(1)
 {
  zn=PRG_RDB(&tekst[nr++]);  //pobranie znaku z pamici programu
  if(zn!=0)                  //czy nie ma koca tekstu?
  {
   if(zn==CR)                //czy znak nowej linii
   {
    wiersz==1?wiersz=0:++wiersz;       //przejd do nowej linii
    kolumna=0;
    lcdxy(wiersz,kolumna);   //ustaw obowizujce po zmianie wsprzdne na LCD
   }
   else
   {
    piszdlcd(zn);            //umie pojedynczy znak tekstu na LCD
   }
  }
  else
  {
   break;                    //zakocz ptl, jeli koniec tekstu
  }
 }
}

//>>>>>>>>>>>>>    Procedury obsugi interfejsu 1-Wire   <<<<<<<<<<<<<<<<<<
void czekaj_1w(unsigned char tau1w)    //ptla opnienia dla 1-wire 
{                       //opnienie jest rwne t=(5+5*tau1w)*T     
 do                     //T - cykl zegarowy MCU, tau1w - parametr procedury
 {
  asm("nop");           //wstawka asemblerowa - rozkaz NOP mikrokontrolera
 }while(--tau1w!=0);
}

unsigned char slot1w_czyt(void)  //slot odczytu pastylki
{
 unsigned char bit1w;

 pastylka_wy=stan_0;         //inicjuj slot czytania
 czekaj_1w(1);               //odczekaj tsu
 pastylka_wy=stan_1;         //zwolnij lini
 czekaj_1w(27);               //ok. 14us
 bit1w=pastylka_we==1?1:0;
 czekaj_1w(118);             //ok. 60us
 return bit1w;
}

void slot1w_zap(unsigned char znak)       //slot zapisu do pastylki
{
 if(znak)
 {                           //slot "1"
  pastylka_wy=stan_0;        //inicjuj slot
  czekaj_1w(20);             //ok. 11us
  pastylka_wy=stan_1;        //podciganie do "1" rezystorem zewntrznym
  czekaj_1w(119);            //60us do koca slotu
 }
 else
 {                           //slot "0"
  pastylka_wy=stan_0;        //inicjuj slot
  czekaj_1w(139);            //ok. 70us
  pastylka_wy=stan_1;        //podciganie do "1" rezystorem zewntrznym
 }
}

void zapisz1w(unsigned char rozkaz)    //transmisja 8-bitowego rozkazu do pastylki
{
 unsigned char i;

 for(i=0;i<8;i++)
 {
  slot1w_zap(rozkaz&0x01); //wylij bit do pastylki
  rozkaz>>=1;
 }
}

void czytaj1w(void)          //odczyt bajtu z pastylki
{
 unsigned char i,j;
 unsigned char dana;

 pbufor1w=&bufor1w[0];  //dane bd umieszczone w buforze "bufor1w"
 for(i=0;i<8;i++)       //czytaj "family code" (1bajt) 
 {                      //i "registration number" (6 bajtw)
  dana=0;               //wstpne zerowanie danej
  for(j=0x01;j!=0;j<<=1)     //zmienna sterujca ptli wskazuje jednoczenie
  {                          //aktualnie zapisywany bit
   dana|=slot1w_czyt()?j:dana;   //czytaj kolejne bity (s one sumowane logicznie
  }                          //z odczytanymi wczeniej
  *pbufor1w++=dana;          //po skompletowaniu zapisz odebrany bajt do bufora
 }
}

void licz_CRC(char bajt,unsigned char *CRC)      //procedura wyliczania CRC
{                  //wielomian generujcy jest rwny: x^8 + x^5 + x^4 + 1
 unsigned char zp1,zp2,i;    //zmienne pomocnicze

 zp1=bajt;
 for(i=0;i<8;i++)
 {
  bajt^=*CRC;           //wskanik *CRC wyznacza aktualnie wyliczony CRC
  zp2=bajt&0x01;        //wydzielenie bitu do oblicze
  bajt=*CRC;
  if(zp2)
  {
   bajt^=0x18;
  }
  bajt=((unsigned char)(bajt)>>1)+0x80*zp2; //konwersja uch jest potrzebna do
                                            //prawidowego wykonania przesunicia 
  *CRC=bajt;
  zp1=(bajt=zp1>>1);    
 }
}

int main(void)     //program gwny
{
 unsigned char i,zp;

 //tablica komunikatw do wywietlenia
 char *info[5]={
                PSTR("Wykryto pastylke"),
                PSTR("Wykryto zwarcie "),
                PSTR("Brak pastylki   "),
                PSTR("                "),
                PSTR("Blad CRC        "),
               };

 DDRB=0xff;        //PORTB - wy
 DDRD=0;           //PORTD - we
 PORTD=0;          //bez podcigania
 led0=1;           //ga LED-a

//>>>>>>>>>>>>>>>>>>    Inicjowanie wywietlacza LCD   <<<<<<<<<<<<<<<<<<<<<<<<<
 lcd_rs=0;
 czekaj(45*tau);        //opnienie ok. 45ms dla ustabilizowania si napicia
                        //zasilania LCD (katalogowo min. 15 ms)
 for(i=0;i<3;i++)       //3-krotne wysanie 3-
 {
  lcd_e=1;
  PORTB=(PORTB&0x0f)|0x30;   //wylij 3- do LCD
  asm("nop");
  asm("nop");
  asm("nop");
  lcd_e=0;
  czekaj(5*tau);        //ok. 5ms
 }
 lcd_e=1;
 PORTB=(PORTB&0x0f)|0x20;    //wylij 2- do LCD
 asm("nop");                 //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 lcd_e=0;               //impuls strobujcy
 czekaj(10L);
 piszilcd(0x28);        //interfejs 4-bitowy, 2 linie, znak 5x7
 piszilcd(0x08);        //wycz LCD, wycz kursor, wycz mruganie
 piszilcd(0x01);        //czy LCD
 czekaj(1.64*tau);      //wymagane dla instrukcji czyszczenia ekranu opnienie
 piszilcd(0x06);        //bez przesuwania w prawo
 piszilcd(0x0c);        //wcz LCD, bez kursora, bez mrugania

 while(1)               //gwna ptla programu
 {
  zp=0;
  pastylka_wy=stan_0;   //sekwencja inicjujca pastylk
  czekaj(5);            //impuls inicjujcy "0" ok. 480us
  pastylka_wy=stan_1;   //zwolnij lini
  czekaj_1w(128);       //po odczekaniu ok. 65us czekaj na impuls obecnoci
  if(pastylka_we==0)
  {
   zp++;                //zapamitaj fakt wykrycia impulsu obecnoci pastylki
  }
  czekaj(4);
  czekaj_1w(56);        //czekaj ok. 416us do zakoczenia slotu inicjujcego
  if(pastylka_we==0)
  {
   zp++;           //jeli pozostaje w stanie niskim, to oznacza zwyke zwarcie
  }
  if(zp==1)        //reakcja na rozpoznane sytuacje
  {                //jeli zp=1, to oznacza e pastylka zgosia si
   zapisz1w(0x33); //wylij do pastylki kod rozkazu przesania numeru seryjnego 
   czytaj1w();     //czytaj numer seryjny pastylki (+ kod rodziny + CRC)
   zp=0;           //w zp bdzie liczone CRC - wstpne zerowanie
   for(i=0;i<8;i++)
   {
    licz_CRC(bufor1w[i],&zp);     //licz CRC ze wszystkich bajtw odebranych
   }
   lcdxy(0,0);
   if(zp==0)
   {               //zerowa warto CRC oznacza prawidowo danych
    led0=0;        //zapal LED-a
    pisztekst(info[0]);      //wywietl komunikat na LCD o rozpoznaniu pastylki
    lcdxy(1,0);
    pbufor1w=&bufor1w[7];
    for(i=6;i!=0xff;i--)     //wywietlenie danych z pastylki
    {
     utoa((unsigned char)bufor1w[i],buflcd,16); //przepisz bufor1w do
                   //bufora wywietlacza z jednoczesn konwersj na ASCII
     pbuflcd=&buflcd[0];     //ustaw wskanik bufora LCD na pocztek 
     if(strlen(buflcd)<2)    //jeli odebrano pojedyncz cyfr, trzeba dopisa 0
     {
      piszznak('0');         //wywietlenie zera wiodcego
     }
     while(*pbuflcd)         //wywietl zawarto bufora LCD
     {
      piszznak(toupper(*pbuflcd++));  //wywietl dane z konwersj na due litery
     }
    }
   }
   else
   {
    pisztekst(info[4]);      //komunikat o bdzie obliczenia CRC
   }
   czekaj(3000*tau);         //przytrzymaj wynik na wywietlaczu przez ok. 3s
  }
  else
  {
   led0=1;                  //ga LED-a
   lcdxy(1,0);
   pisztekst(info[3]);      //wyczy doln lini wywietlacza
   lcdxy(0,0);
   if(zp==2)
   {                         //to byo tylko zwarcie
    pisztekst(info[1]);      //wywietl komunikat o zwarciu
   }
   else
   {
    pisztekst(info[2]);      //wywietl komunikat o braku pastylki
   }
  }
 }
}
