/*********************************************************************/
/* wiczenie 10 - Obsuga interfejsu I2C                             */
/*                ukad RTC (zegar czasu rzeczywistego) - PCF8583    */
/*                Obsuga przerwania zewntrznego INT0               */
/*                Obsuga wywietlacza LCD 2x16                       */
/* J.D. '2003                                                        */
/*********************************************************************/

#include <io2313.h>
#include <progmem.h>
#include <interrupt.h>
#include <signal.h>

//  Ponisze definicje su do realizacji wygodnego dostpu bitowego
typedef struct _bit_struct
{
 unsigned char bit0: 1;
 unsigned char bit1: 1;
 unsigned char bit2: 1;
 unsigned char bit3: 1;
 unsigned char bit4: 1;
 unsigned char bit5: 1;
 unsigned char bit6: 1;
 unsigned char bit7: 1;
}pole_bitowe;

#define DAJ_BIT(adr) (*((volatile pole_bitowe*) (adr)))
#define _PORTB 0x38
#define _PINB 0x36
#define _PORTD 0x32
#define _DDRD 0x31
#define _PIND 0x30
#define sda_we DAJ_BIT(_PIND).bit5
#define sda_wy DAJ_BIT(_DDRD).bit5
#define scl_wy DAJ_BIT(_DDRD).bit6
#define scl_we DAJ_BIT(_PIND).bit6
#define lcd_rs DAJ_BIT(_PORTB).bit2
#define lcd_e DAJ_BIT(_PORTB).bit3
#define sw4 DAJ_BIT(_PIND).bit1
#define sw1 DAJ_BIT(_PIND).bit0

#define stan_0 1   //definicja stanu niskiego na liniach I2C
                   //"1" oznacza przeczenie portu w tryb wyjciowy
                   //port jest wczeniej wysterowany w stan niski
#define stan_1 0   //definicja stanu wysokiego na liniach I2C
                   //"0" oznacza przeczenie portu w tryb wejciowy
                   //stan wysoki jest wymuszany przez zewntrzny
                   //rezystor podcigajcy

#define rtc 0xa0   //7-bitowy adres bazowy zegara RTC (PCF8583) = 1010000
                   //przesunity na bity od 7 do 1 i uzupeniony zerem
#define CR 0x0a    //definicja znaku CR (przecie do nowej linii)

char buflcd[4];    //roboczy bufor wywietlacza LCD
char *pbuflcd;     //wskanik na bufor wywietlacza
unsigned char bufi2c[9];     //bufor interfejsu I2C (dane z/do RTC)
unsigned char *pbufi2c;      //wskanik na bufor danych z RTC
unsigned char wiersz=0;      //pozycja umieszczenia znaku na LCD
unsigned char kolumna=0;     //pozycja umieszczenia znaku na LCD
unsigned char godz=0,min=0,sek=0,dzien=1,mies=1;
unsigned char fzegar;        //flaga odwieenia stanu zegara na LCD

void czekaj(unsigned long zt)     //procedura wytracania czasu
{
 #define tau 10.38           //przybliony przelicznik argumentu na ms
 unsigned char zt1;
 for(;zt>0;zt--)
 {
  for(zt1=255;zt1!=0;zt1--);
 }
}

//>>>>>>>>>>>>>    Procedury obsugi wywietlacza LCD   <<<<<<<<<<<<<<<<<<
void piszilcd(unsigned char instr)     //zapisz instrukcj sterujc do LCD
{
 lcd_rs=0;
 lcd_e=1;
 PORTB=(PORTB&0x0f)|(instr&0xf0);      //przygotuj starszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD ok. 100us
 lcd_e=1;
 PORTB=(PORTB&0x0f)|((instr&0x0f)<<4); //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD ok. 100us
}

void piszdlcd(char dana)     //zapisz dan do LCD
{
 lcd_rs=1;
 lcd_e=1;
 PORTB=(PORTB&0x0f)|(dana&0xf0);  //przygotuj starszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD
 lcd_e=1;
 PORTB=(PORTB&0x0f)|((dana&0x0f)<<4);  //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD
}

void czysclcd(void)          //czy ekran
{
 piszilcd(0x01);   //polecenie czyszczenia ekranu dla kontrolera LCD
 czekaj(1.64*tau); //rozkaz 0x01 wykonuje si 1.64ms
 wiersz=0;
 kolumna=0;
}

void lcdxy(unsigned char w, unsigned char k)     //ustaw wsprzdne kursora
{
 piszilcd((w*0x40+k)|0x80);  //standardowy rozkaz sterownika LCD
}                            //ustawiajcy kursor w okrelonych wsprzdnych

void piszznak(char znak)     //procedura umieszcza znak na wywietlaczu
{
 piszdlcd(znak);             //wywietl znak na LCD
}

void pisztekst(char *tekst)  //pisz tekst na LCD wskazywany pointerem *tekst
{
 char zn;
 char nr=0;

 while(1)
 {
  zn=PRG_RDB(&tekst[nr++]);  //pobranie znaku z pamici programu
  if(zn!=0)                  //czy nie ma koca tekstu?
  {
   if(zn==CR)                //czy znak nowej linii
   {
    wiersz==1?wiersz=0:++wiersz;       //przejd do nowej linii
    kolumna=0;
    lcdxy(wiersz,kolumna);   //ustaw obowizujce po zmianie wsprzdne na LCD
   }
   else
   {
    piszdlcd(zn);            //umie pojedynczy znak tekstu na LCD
   }
  }
  else
  {
   break;                    //zakocz ptl, jeli koniec tekstu
  }
 }
}

void nalcd(unsigned char zn1, unsigned char zn2)
{                  //umieszczenie czasu na LCD z jednoczesn konwersj na ASCII
 piszznak(zn1+0x30);    //dodanie 0x30 realizuje prost konwersj liczby na ASCII
 piszznak((zn2&0x0f)+0x30);
}

//>>>>>>>>>>>>>>>>      Procedury obsugi interfejsu I2C    <<<<<<<<<<<<<<<<<<<<<
void czekaj_i2c(unsigned char taui2c)  //ptla opnienia dla I2C
{                       //opnienie jest rwne t=(5+5*taui2c)*T
 do                     //T - cykl zegarowy MCU, taui2c - parametr procedury
 {
  asm("nop");           //wstawka asemblerowa - rozkaz NOP mikrokontrolera
 }while(--taui2c!=0);
}

void bitstartu(void)    //bit startu na magistrali I2C
{                       //     SCL    SDA
 sda_wy=stan_1;         //               |
 czekaj_i2c(10);        //               |
 scl_wy=stan_1;         //        |      |
 czekaj_i2c(10);        //        |      |
 sda_wy=stan_0;         //        |     /
 czekaj_i2c(10);        //        |    |
 scl_wy=stan_0;         //        /    |
 czekaj_i2c(10);        //       |     |
}                       //       |     |

void bitstopu(void)
{                       //     SCL    SDA
 sda_wy=stan_0;         //     |       |
 czekaj_i2c(10);        //     |       |
 scl_wy=stan_1;         //      \      |
 czekaj_i2c(10);        //       |     |
 sda_wy=stan_1;         //       |      \
 czekaj_i2c(10);        //       |       | 
}                       //       |       |

unsigned char zapiszB_i2c(unsigned char dana)
{                       //wysanie pojedynczego bajtu do Slave'a I2C
 unsigned char i;

 for(i=0;i<8;i++)       //bdzie 8 bitw 
 {
  if(dana&0x80)         //badaj najstarszy bit wysyanego znaku
  {
   sda_wy=stan_1;       //wylij "1"
  }
  else
  {
   sda_wy=stan_0;       //wylij "0"
  }
  czekaj_i2c(10);
  scl_wy=stan_1;
  czekaj_i2c(10);
  scl_wy=stan_0;
  czekaj_i2c(10);
  dana<<=1;        //przygotuj nastpny bit do wysania
 }
 sda_wy=stan_1;    //zwolnij lini SDA dla Slave'a
 czekaj_i2c(10);
 scl_wy=stan_1; 
 while(!scl_we);   //czekaj, a Slave bdzie gotowy do przyjcia nastpnego bitu
 czekaj_i2c(10);
 i=sda_we;         //czytaj potwierdzenie ACK od Slave'a
 scl_wy=stan_0;
 czekaj_i2c(10);
 return i;         //zwr bit potwierdzenia ACK
}

unsigned char czytajB_i2c(unsigned char ack)
{        //przyjcie pojedynczego bajtu od Slave'a I2C
         //parametr ack=1 oznacza, e jest wysyany ostatni bajt bloku,
         //nie naley wic wysya potwierdzenia
 unsigned char dana=0,i;

 for(i=0;i<8;i++)  //bdzie 8 bitw 
 {
  scl_wy=stan_0;
  czekaj_i2c(10);
  scl_wy=stan_1;
  czekaj_i2c(10);
  dana<<=1;             //przygotuj miejsce na kolejny odebrany bit
  dana|=sda_we==1?1:0;  //dopisz odebrany bit  
 }
 scl_wy=stan_0;
 if(ack)           //czy wysya potwierdzenie?
 {
  sda_wy=stan_0;   //wylij ACK
 }
 else
 {
  sda_wy=stan_1;   //brak potwierdzenia sygnalizuje zakoczenie bloku
 }
 czekaj_i2c(10);
 scl_wy=stan_1;
 while(!scl_we);   //czekaj, a Slave bdzie gotowy do przyjcia nastpnego bitu
 czekaj_i2c(10);
 sda_wy=stan_1;    //zwolnij lini SDA
 scl_wy=stan_0;
 return dana;
}

void bladi2c(void)
{                  //reakcja na brak potwierdzenia ACK od Slave'a
 char *tekst[2]={
                PSTR("Blad I2C"),
                PSTR("        ")
               };
 lcdxy(0,0);
 pisztekst(tekst[0]);   //wywietl na LCD komunikat o bdzie I2C
 czekaj(1500*tau);
 lcdxy(0,0);
 pisztekst(tekst[1]);   //wyczy komunikat o bdzie
}

void doi2c(unsigned char adri2c,unsigned char adrdane,unsigned char lz)
{        //wysanie bloku "lz" znakw do Slave'a I2C
         //"adri2c" - adres fizyczny Slave'a
         //"adrdane" - adres rejestru w Slave'ie, od ktrego bd zapisywane dane
 pbufi2c=&bufi2c;  //ustaw wskanik bufora i2c na jego pocztek
 bitstartu();      //wylij bit startu (pocztek transmisji)
 if(zapiszB_i2c(adri2c))     //wylij adres Slave'a, to informacja dla pozostaych 
                             //ukadw Slave, e dane nie bd dla nich
 {
  bladi2c();  //wywietl komunikat o bdzie interfejsu I2C, jeli nie byo ACK
 }
 if(zapiszB_i2c(adrdane))    //zaadresuj rejestr w Slave'ie
 {
  bladi2c();  //wywietl komunikat o bdzie interfejsu I2C, jeli nie byo ACK
 }
 for(;lz!=0;lz--)       //wysyanie bloku danych (lz - liczba bajtw do wysania
 {
  if(zapiszB_i2c(*pbufi2c++))     //wylij kolejny bajt do Slave'a
  {
   bladi2c(); //wywietl komunikat o bdzie interfejsu I2C, jeli nie byo ACK
  }
 }
 bitstopu();       //wylij bit stopu (koniec transmisji)
}

void odi2c(unsigned char adri2c,unsigned char adrdane,unsigned char lz)
{        //odebranie bloku "lz" znakw od Slave'a I2C
         //"adri2c" - adres fizyczny Slave'a
         //"adrdane" - adres rejestru w Slave'ie, od ktrego bd odczytywane dane
 bitstartu();      //wylij bit startu (pocztek transmisji)
 if(zapiszB_i2c(adri2c))     //wylij adres Slave'a, to informacja dla wybranego
                             //Slave'a, e dane bd odbierane od niego
 {
  bladi2c();  //wywietl komunikat o bdzie interfejsu I2C, jeli nie byo ACK
 }
 if(zapiszB_i2c(adrdane))    //zaadresuj rejestr w Slave'ie
 {
  bladi2c();  //wywietl komunikat o bdzie interfejsu I2C, jeli nie byo ACK
 }
 bitstartu();      //ponowne wysanie bitu startu 
 if(zapiszB_i2c(adri2c+1))   //ponowne wysanie adresu Slave'a z bitem R/W=1
                             //oznacza przeczenie Slave'a na nadawanie
 {               
  bladi2c();  //wywietl komunikat o bdzie interfejsu I2C, jeli nie byo ACK
 }
 pbufi2c=&bufi2c;  //ustaw wskanik bufora i2c na jego pocztek
 for(;lz>1;lz--)   //odbierz "lz" bajtw danych od Slave'a
 {
  *pbufi2c++=czytajB_i2c(1); //zapisuj odebrane bajty w buforze i2c
 }
 *pbufi2c++=czytajB_i2c(0);  //odbierz ostatni bajt od I2C, nie wysyaj ACK
 bitstopu();       //wylij bit stopu  (koniec transmisji)
}

SIGNAL(SIG_INTERRUPT0)       //procedura obsugi przerwania zewntrznego INT0
{
 PORTB^=0x03;                //zmie stany LED1 i LED2
 odi2c(rtc,0x02,3);          //odczytaj czas z RTC (tylko godz, min i sek)
 fzegar=1;                   //ustaw flag odczytania danych 
}

void wyswietlczas(void)      //procedura wywietlania czasu na LCD
{
 unsigned char zp;           //zmienna pomocnicza

 lcdxy(1,0);
 pbufi2c=&bufi2c[2];         //ustaw wskanik bufora na pozycje godzin
 zp=*pbufi2c--;              //pobierz godziny
 nalcd((zp&0x30)>>4,zp);     //wywietl godziny
 piszznak(':');
 zp=*pbufi2c--;              //pobierz minuty
 nalcd((zp&0xf0)>>4,zp);     //wywietl minuty
 piszznak(':');
 zp=*pbufi2c;                //pobierz sekundy
 nalcd((zp&0xf0)>>4,zp);     //wywietl minuty
}

void zwolnijklaw(void)      
{
 czekaj(200*tau);
 while((PIND&0x03)==0x03);   //czekaj a wszystkie klawisze bd zwolnione
}

void ustawzegar(void)        //procedura ustawiania czasu w ukadzie RTC
{
 char *kom[1]={PSTR("STARTn/t")};

 cli();                 //wycz przerwania, aby nie naczyta danych
 piszilcd(0x0d);        //wcz mruganie
 odi2c(rtc,0x02,3);     //odczytaj czas z RTC (tylko godz, min, sek)
 wyswietlczas();
 pbufi2c=&bufi2c[1];    //ustaw wskanik bufora i2c na minuty
 bufi2c[0]=0;           //sekundy bd zawsze zerowane
 min=*pbufi2c++;        //pobierz minuty
 godz=*pbufi2c;         //pobierz godziny
 lcdxy(1,1);
 zwolnijklaw();
 while(sw4)        //czekaj na nacinicie SW4 wykonujc ponisze instrukcje
 {                 //ustawianie godzin
  lcdxy(1,1);
  if(!sw1)         //czy nacinito SW1?
  {
   godz++;         //inkrementuj godziny
   if((godz&0x0f)>9)
   {
    godz+=0x06;    //korekcja dziesitna liczby BCD
   }
   if((godz&0x3f)>0x23)
   {
    godz&=0xc0;         //kasuj godziny po przekroczeniu zakresu
   }
   bufi2c[2]=(bufi2c[2]&0xc0)|godz;    //zapisz uaktualnione godziny w buforze i2c
                        //pozostawiajc bity 24/12 i AM/PM nienaruszone
   wyswietlczas();
   czekaj(150*tau);
  }
 }
 lcdxy(1,4);
 zwolnijklaw();

 while(sw4)        //czekaj na nacinicie SW4 wykonujc ponisze instrukcje
 {                 //ustawianie minut
  lcdxy(1,4);
  if(!sw1)         //czy nacinito SW1?
  {
   min++;          //inkrementuj godziny
   if((min&0x0f)>9)
   {
    min+=0x06;     //korekcja dziesitna liczby BCD
   }
   if(min>0x59)    //korekcja przekroczenia wartoci 59
   {
    min&=0;   //przeniesienia na godziny nie ma, bo s one ustawiane niezalenie
   }
   bufi2c[1]=min;  //zapisz uaktualnione minuty w buforze i2c
   wyswietlczas();
   czekaj(150*tau);
  }
 }
 piszilcd(0x0c);        //wycz mruganie
 lcdxy(1,8);
 pisztekst(kom[0]);     //wywietl komunikat "START"
 zwolnijklaw();
 while(sw1&sw4);        //czekaj na nacinicie jakiego klawisza
 if(!sw4)               //po naciniciu klawisza SW4...
 {
  doi2c(rtc,2,3);        //...zapisz ustawienia zegara
  bufi2c[0]=0;
  doi2c(rtc,0,1);        //start zegara
 }
 sei();            //wcz przerwania
}

int main(void)     //program gwny
{
 unsigned char i,zp;

 //tablica komunikatw do wywietlenia
 char *info[4]={
                PSTR("Ustaw zegar     "),
                PSTR("Blad I2C        "),
                PSTR("RTC  -  PCF8583 "),
                PSTR("ZL1AVR - plytka\newaluacyjna AVR")
               };
 DDRB=0xff;        //PORTB - wy
 PORTB|=0x02;      //zapal LED1, zga LED2
 DDRD=0x00;        //PORTD - we
 PORTD=0x07;       //podciganie linii PD0, PD1 i PD2
 scl_wy=stan_0;    //SCL=0
 fzegar=0;

//>>>>>>>>>>>>>>>>>>    Inicjowanie wywietlacza LCD   <<<<<<<<<<<<<<<<<<<<<<<<<
 lcd_rs=0;
 czekaj(45*tau);        //opnienie ok. 45ms dla ustabilizowania si napicia
                        //zasilania LCD (katalogowo min. 15 ms)
 for(i=0;i<3;i++)       //3-krotne wysanie 3-
 {
  lcd_e=1;
  PORTB=(PORTB&0x0f)|0x30;   //wylij 3- do LCD
  asm("nop");
  asm("nop");
  asm("nop");
  lcd_e=0;
  czekaj(5*tau);        //ok. 5ms
 }
 lcd_e=1;
 PORTB=(PORTB&0x0f)|0x20;    //wylij 2- do LCD
 asm("nop");                 //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 lcd_e=0;               //impuls strobujcy
 czekaj(10L);
 piszilcd(0x28);        //interfejs 4-bitowy, 2 linie, znak 5x7
 piszilcd(0x08);        //wycz LCD, wycz kursor, wycz mruganie
 piszilcd(0x01);        //czy LCD
 czekaj(1.64*tau);      //wymagane dla instrukcji czyszczenia ekranu opnienie
 piszilcd(0x06);        //bez przesuwania w prawo
 piszilcd(0x0c);        //wcz LCD, bez kursora, bez mrugania
//>>>>>>>>>>>>>>>>>>               <<<<<<<<<<<<<<<<<<<<<<<<<

 GIMSK=1<<INT0;         //wcz przerwanie zewntrzne INT0
 MCUCR=0x02;            //przerwania INT0 wyzwalane opadajcym zboczem
 if(!sw1)               //czy w chwili restartu by nacinity SW1?
 {
  czysclcd();
  pisztekst(info[3]);   //wywietl info o pytce
  czekaj(1000*tau);
  while(sw1);           //czekaj na ponowne nacinicie SW1
  czysclcd();
 }
 pisztekst(info[2]);    //wywietl komunikat o ukadzie RTC na LCD
 sei();                 //wcz przerwania
 while(1)
 {
  if(fzegar)            //czy mona odwiey czas na LCD
  {
   fzegar=0;
   lcdxy(1,0);
//   odi2c(rtc,0x02,3);   
   wyswietlczas();
  }
  if(!sw4)              //czy nacinito klawisz SW4
  {
   czekaj(150*tau);
   lcdxy(0,0);
   pisztekst(info[0]);  //wywietl komunikat o trybie ustawiania
   ustawzegar();        //ustaw zegar
   czysclcd();
   pisztekst(info[2]);  //odtwrz komunikat o ukadzie RTC na LCD
   czekaj(300*tau);
  }
 }
}

