/********************************************************************/
/* wiczenie 11 - Poczenie mikrokontrolera AVR do komputera PC    */
/*                poprzez port USB                                  */
/*                Nadawanie i odbir poprzez UART z uyciem przerwa*/
/* J.D. '2003                                                       */
/********************************************************************/

#include <io.h>
#include <progmem.h>
#include <stdlib.h>
#include <interrupt.h>
#include <signal.h>

#define FCPU      8000000    //czstotliwo oscylatora CPU
#define VUART     38400      //prdko transmisji [b/s]
#define VUBRR     FCPU/(VUART*16)-1    //wpis do UBRR dla VUART

//  Ponisze definicje su do realizacji wygodnego dostpu bitowego
typedef struct _bit_struct
{
 unsigned char bit0: 1;
 unsigned char bit1: 1;
 unsigned char bit2: 1;
 unsigned char bit3: 1;
 unsigned char bit4: 1;
 unsigned char bit5: 1;
 unsigned char bit6: 1;
 unsigned char bit7: 1;
}pole_bitowe;

#define DAJ_BIT(adr) (*((volatile pole_bitowe*) (adr)))
#define _PORTB 0x38
#define _PINB 0x36
#define _PORTD 0x32
#define _DDRD 0x31
#define _PIND 0x30
#define lcd_rs DAJ_BIT(_PORTB).bit2
#define lcd_e DAJ_BIT(_PORTB).bit3
#define led0 DAJ_BIT(_PORTB).bit0
#define led1 DAJ_BIT(_PORTB).bit1
#define LF 0x0a    //definicja znaku LF (przecie do nowej linii)
#define CR 0x0d    //definicja znaku CR (powrt karetki)

unsigned char wiersz=0;      //pozycja umieszczenia znaku na LCD
unsigned char kolumna=0;     //pozycja umieszczenia znaku na LCD

char volatile iofifosio;     //wskanik odczytu kolejki UART-u
char volatile izfifosio;     //wskanik zapisu kolejki UART-u
char fifosio[32];            //kolejka UART-u
unsigned char volatile ldanych=0;      //liczba danych w buforze fifosio

void czekaj(unsigned long zt)     //procedura wytracania czasu
{
 #define tau 10.38           //przybliony przelicznik argumentu na ms
 unsigned char zt1;
 for(;zt>0;zt--)
 {
  for(zt1=255;zt1!=0;zt1--);
 }
}

//>>>>>>>>>>>>>    Procedury obsugi wywietlacza LCD   <<<<<<<<<<<<<<<<<<
void piszilcd(unsigned char instr)     //zapisz instrukcj sterujc do LCD
{
 lcd_rs=0;
 lcd_e=1;
 PORTB=(PORTB&0x0f)|(instr&0xf0);  //przygotuj starszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD ok. 100us
 lcd_e=1;
 PORTB=(PORTB&0x0f)|((instr&0x0f)<<4); //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD ok. 100us
}

void piszdlcd(char dana)     //zapisz dan do LCD
{
 lcd_rs=1;
 lcd_e=1;
 PORTB=(PORTB&0x0f)|(dana&0xf0);  //przygotuj starszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD
 lcd_e=1;
 PORTB=(PORTB&0x0f)|((dana&0x0f)<<4);  //przygotuj modszy pbajt do LCD
 asm("nop");
 asm("nop");
 asm("nop");
 lcd_e=0;          //impuls strobujcy
 czekaj(10L);      //czekaj na gotowo LCD
}

void czysclcd(void)          //czy ekran
{
 piszilcd(0x01);   //polecenie czyszczenia ekranu dla kontrolera LCD
 czekaj(1.64*tau); //rozkaz 0x01 wykonuje si 1.64ms
 wiersz=0;
 kolumna=0;
}

void lcdxy(unsigned char w, unsigned char k)     //ustaw wsprzdne kursora
{
 piszilcd((w*0x40+k)|0x80);  //standardowy rozkaz sterownika LCD
}                            //ustawiajcy kursor w okrelonych wsprzdnych

void piszznak(char znak)     //procedura umieszcza znak na wywietlaczu
{
 piszdlcd(znak);             //wywietl znak na LCD
}

void pisztekst(char *tekst,unsigned char romram) //pisz tekst na LCD wskazywany
                                                 //pointerem *tekst
{
 char zn;
 char nr=0;

 while(1)
 {
  if(!romram)
  {
   zn=PRG_RDB(&tekst[nr++]);      //pobranie znaku z pamici programu
  }
  else
  {
   if(!ldanych)              //czy s jeszcze jakie dane w buforze
   {
    break;
   }
   zn=fifosio[iofifosio];    //pobierz znak z bufora
   iofifosio==31?iofifosio=0:++iofifosio;   //inkrementacja modulo 32
   ldanych--;
  }
  if(zn!=0)                  //czy nie ma koca tekstu?
  {
   if((zn==LF)||(zn==CR))    //czy znak nowej linii?
   {

    if(zn==LF)
    {
     wiersz==1?wiersz=0:++wiersz;      //przejd do nowej linii
    }
    else           //by CR
    {
     kolumna=0;    //powrt karetki 
    }
   }
   else
   {
    piszdlcd(zn);            //umie pojedynczy znak tekstu na LCD
    kolumna==15?kolumna=0:++kolumna;   //inkrementacja modulo 16
    if(kolumna==0)
    {
     wiersz==1?wiersz=0:++wiersz;      //przejd do nowej linii
    }
   }
   lcdxy(wiersz,kolumna);
  }
  else
  {
   break;                    //zakocz ptl, jeli koniec tekstu
  }
 }
}

SIGNAL(SIG_UART_RECV)        //procedura obsugi odbiornika UART-u
{
 led1=0;
 if(ldanych<32)
 {
  fifosio[izfifosio]=UDR;                   //zapamitaj odebrany znak
  izfifosio==31?izfifosio=0:++izfifosio;    //inkrementacja modulo 32
  ldanych++;
 }
 led1=1;                     //zga leda
}

SIGNAL(SIG_UART_TRANS)       //procedura obsugi nadajnika UART
{                            //wywoywana po wysaniu znaku
 char znak;

 led0=0;
 znak=fifosio[iofifosio];    //pobierz dane z pamici RAM
 if(ldanych!=0)              //czy koniec pobierania danych?
 {
  UDR=znak;                  //nie, wylij znak pobrany z kolejki
  ldanych--;
 }
 else
 {
  cbi(UCR,TXEN);             //tak, wycz nadajnik
 }
 iofifosio==31?iofifosio=0:++iofifosio;     //inkrementacja mod 32
 led0=1;
}

int main(void)
{
 unsigned long i;
 char znak;

 //tablica komunikatw do wysania
 char *info[2]={
                PSTR("Wysylam dane"),
                PSTR("Przeslij z PC-ta")
               };

 DDRB=0xff;    //PORTB - wy
 PORTB=0xff;
 DDRD=0x02;    //PD1 - wy (RXD), pozostae we
 PORTD=0x02;   //podcigania wejcia PD1 (RXD)

 lcd_rs=0;
 czekaj(45*tau);        //opnienie ok. 45ms dla ustabilizowania si napicia
                        //zasilania LCD (katalogowo min. 15 ms)
 for(i=0;i<3;i++)       //3-krotne wysanie 3-
 {
  lcd_e=1;
  PORTB=(PORTB&0x0f)|0x30;   //wylij 3- do LCD
  asm("nop");
  asm("nop");
  asm("nop");
  lcd_e=0;
  czekaj(5*tau);        //ok. 5ms
 }
 lcd_e=1;
 PORTB=(PORTB&0x0f)|0x20;    //wylij 2- do LCD
 asm("nop");                 //wymagane wyduenie impulsu
 asm("nop");
 asm("nop");
 lcd_e=0;               //impuls strobujcy
 czekaj(10L);
 piszilcd(0x28);        //interfejs 4-bitowy, 2 linie, znak 5x7
 piszilcd(0x08);        //wycz LCD, wycz kursor, wycz mruganie
 piszilcd(0x01);        //czy LCD
 czekaj(1.64*tau);      //wymagane dla instrukcji czyszczenia ekranu opnienie
 piszilcd(0x06);        //bez przesuwania w prawo
 piszilcd(0x0c);        //wcz LCD, bez kursora, bez mrugania

 pisztekst(info[0],0);  //komunikat na LCD
 UBRR=VUBRR;            //ustaw prdko transmisji
 UCR=1<<RXCIE | 1<<TXCIE  | 1<<RXEN;   //zezwolenie na przerwania od
                   //odbiornika i nadajnika, zezwolenie na odbir i nadawanie
 sei();            //wcz przerwania

 izfifosio=0;      //inicjuj zmienne
 iofifosio=0;
 znak=' ';
 sbi(UCR,TXEN);         //wcz nadajnik
 UDR=znak++;            //umie pierwszy znak
 for(i=0;i<0x1ffff;i++) //ptla wysyania znakw
 {
  while(ldanych==32);   //jeli bufor przepeniony, to czekaj
  fifosio[izfifosio]=znak++;      //umie znak do wysania w buforze
  if(znak>'Z')
  {
   znak=' ';
  }
  izfifosio==31?izfifosio=0:++izfifosio;    //inkrementuj mod 32
  ldanych++;
 }
 czysclcd();
 pisztekst(info[1],0);
 czekaj(2000*tau);
 czysclcd();

 izfifosio=0;      //inicjuj zmienne
 iofifosio=0;
 ldanych=0;
 while(1)          //ptla odbioru znakw
 {
  if(ldanych)      //czy odebrano jaki znak?
  {
   pisztekst(fifosio[0],1);  //wywietl odebrane znaki
  }
 }
}
